@@ -15,11 +15,11 @@ import SCons.Script
 from SCons.Script import AddOption, GetOption, SetOption
 from SCons.Builder import Builder
 
-pre_sys_stdout = sys.stdout #Need this sometimes to print while run from Stata (e.g., stata_run)
-
-if packaging.version.parse(SCons.__version__) < packaging.version.parse("4.2.0"):
-    SCons.Debug.Trace("WARNING: You are running SCons version" + SCons.__version__ + " and statacons has only been tested on 4.2.0+. \n", pre_sys_stdout)
+pre_sys_stdout = sys.stdout #Need this sometimes to print while run from Stata once building (e.g., stata_run)
 
+##############################################
+#               Configuration                #
+##############################################
 def _find(pathname, paths=None):
     if paths is None:
         paths = os.environ['PATH'].split(os.pathsep)
@@ -60,30 +60,16 @@ AddOption(
     dest='config_user',
     action='store',
     type="string",
-    help='config user',
-)
-if GetOption("config_user") is not None:
-    config = configuration(config_user_ini=GetOption("config_user"))
-else:
-    config = configuration()
-
-AddOption(
-    '--show-config',
-    dest='show_config',
-    action='store_true',
-    default=False,
-    help='show config',
+    help='Specify config_user file path.',
 )
 
-
-
-
-env = DefaultEnvironment(ENV=os.environ, tools=[])
-env['CONFIG'] = config
+##############################################
+#           Stata tool (find it)             #
+##############################################
 def stata_tool(env):
     import platform
     plat = platform.system()
-    if 'Programs' in env['CONFIG'] and 'stata_exe' in env['CONFIG']['Programs']:
+    if 'CONFIG' in env and 'Programs' in env['CONFIG'] and 'stata_exe' in env['CONFIG']['Programs']:
         env['STATABATCHEXE'] = env['CONFIG']['Programs']['stata_exe']
     
     else:      
@@ -102,7 +88,7 @@ def stata_tool(env):
                 break
         if found_ret is None and plat=='Windows':
             for candidate_stata_exe in candidate_stata_exes:
-                found_ret = _find(candidate_stata_exe, [os.path.join(os.environ['ProgramFiles'], "Stata" + str(n)) for n in [17, 16]])
+                found_ret = _find(candidate_stata_exe, [os.path.join(os.environ['ProgramFiles'], "Stata" + str(n)) for n in range(17, 12, -1)])
                 if found_ret is not None:
                     env['STATABATCHEXE'] = found_ret[1]
                     break
@@ -115,6 +101,7 @@ def stata_tool(env):
             
         if found_ret is None:
             print("Can't find Stata from config or defaults")
+            return
 
     batch_flags = {'Windows': '/e', 
                    'Linux': '-b',
@@ -123,29 +110,12 @@ def stata_tool(env):
     env['STATABATCHFLAG'] = batch_flags[plat]
     env['STATABATCHCOM'] = '"' + env['STATABATCHEXE'] + '"' if " " in env['STATABATCHEXE'] else env['STATABATCHEXE']
     env['STATABATCHCOM'] = env['STATABATCHCOM']  + " " + batch_flags[plat]
-   
-env.Tool(stata_tool)
-
-# add the Builder stata_bld to env, give it the name StataBuild
-# tell scons how to tell stata to run a do-file in batch mode
-
-AddOption(
-    '--assume-built',
-    dest='assume_built',
-    action='store',
-    type="string",
-    help='assume built',
-)
-AddOption(
-    '--assume-done',
-    dest='assume_done',
-    action='store',
-    type="string",
-    help='assume done',
-)
 
 
-# Deciders
+
+##############################################
+#              Deciders                      #
+##############################################
 # Note that timestamp-newer/make compares target vs source, whereas timestamp-changed doesn't look at target (just prev vs current of source)
 # 'content-timestamp' = 'timestamp-changed' and 'content'
 
@@ -162,9 +132,27 @@ decider_str_lookup = {'content-timestamp-newer': dependency_newer_then_content_c
 decider_str_lookup.update({k:k for k in ['MD5', 'content', 'MD5-timestamp', 'content-timestamp', 'timestamp-newer', 'make', 'timestamp-match']})
 
 
+##############################################
+#              Builders                      #
+##############################################
 #We impliment these skip-logic routines in stata_run() and not a dedicated Decider so that
+# we can have them thought of as 'built' and info saved in the sconsign file. Also,
 # someone can change them on the command-line without changing any files and is orthogonal
 # to what decider they choose (MD5, MD5-timestamp)
+AddOption(
+    '--assume-built',
+    dest='assume_built',
+    action='store',
+    type="string",
+    help='Assume these targets are up-to-date (and if all targets for a task are up-to-date it will be skipped).',
+)
+AddOption(
+    '--assume-done',
+    dest='assume_done',
+    action='store',
+    type="string",
+    help='Assume these do-files are up-to-date and do no rebuild.',
+)
 assume_built = GetOption('assume_built') #If there are outer quotes on the cli, these are stripped.
 assume_built_list = []
 if not assume_built is None:
@@ -175,49 +163,36 @@ assume_done_list = []
 if not assume_done is None:
     assume_done_list = [str(path) for file_pattern in assume_done.split(':') for path in Path().rglob(file_pattern) if not path.is_dir()] #could've used os.pathsep, but want consistent across sytems
 
-#AddOption(
-#    '--skip-newer',
-#    dest='skip_newer',
-#    action='store_true',
-#    default=False,
-#    help='skip newer',
-#)
-
-# Can't use the Builder chdir option, since that moves the target dir (and we want source)
-stata_chdir = config['SCons']['stata_chdir']
-if stata_chdir=='': stata_chdir = 0
-if stata_chdir=='1' or stata_chdir=='0': stata_chdir = int(stata_chdir)
-
 
+# Handles log file (better than Clean() as we typically want them removed right away (there may be a lot and this is easier))
 def stata_run(target, source, env, params="", file_cmd="do", full_cmd=None):
+    if 'STATABATCHEXE' not in env:
+        raise LookupError("Can't find Stata")
     if full_cmd is None: 
         fname = str(source[0]) #Assumes the first element of source is the do file
         
-    assume_built = GetOption('assume_built')
     if not assume_built is None and all([str(t) in assume_built_list for t in target]):
         if full_cmd is None: full_cmd = file_cmd + ' "' + fname + '" ' + params
         for t in target: Path(str(t)).touch() #In case someone is using a time-stamp option for content being up-to-date. Touch() doesn't seem to work
         if not GetOption('silent'):
             SCons.Debug.Trace("Assuming built: " + full_cmd + "\n", pre_sys_stdout)
         return 0
-    assume_done = GetOption('assume_done')
     if not assume_done is None and 'fname' in locals() and fname in assume_done_list:
         for t in target: Path(str(t)).touch() #In case someone is using a time-stamp option for content being up-to-date. Touch() doesn't seem to work
         if not GetOption('silent'):
             SCons.Debug.Trace("Assuming done: " + fname + "\n", pre_sys_stdout)
         return 0
 
-    #if GetOption('skip_newer') and min([os.path.getmtime(str(t)) for t in target]) > max([os.path.getmtime(str(s)) for s in source]):
-    #    if not GetOption('silent'):
-    #        if full_cmd is None: full_cmd = file_cmd + ' "' + fname + '" ' + params
-    #        SCons.Debug.Trace("Skipping newer: " + full_cmd + "\n", pre_sys_stdout)
-    #    return 0
-
     cwd = None
-    if stata_chdir == 1 and full_cmd is None:
+    
+    # Can't use the Builder chdir option, since that moves the target dir (and we want source)
+    stata_chdir = env['CONFIG']['SCons']['stata_chdir']
+    if stata_chdir=='': stata_chdir = '0'
+
+    if stata_chdir == '1' and full_cmd is None:
         cwd = os.path.dirname(fname)
         fname = os.path.basename(fname)
-    elif stata_chdir != 0:
+    elif stata_chdir != '0':
         cwd = stata_chdir
         fname = os.path.relpath(fname, stata_chdir)
 
@@ -280,8 +255,7 @@ def stata_run(target, source, env, params="", file_cmd="do", full_cmd=None):
         os.replace(log_name, os.path.join(success_batch_log_dir, recipe_basename + ".log"))
 
     return 0
-stata_do_bld = Builder(action=stata_run)
-env.Append(BUILDERS={'StataDo': stata_do_bld})
+    
 
 def copy_func(f):
     import types
@@ -329,40 +303,45 @@ def stata_run_params_factory(self, target, source=None, do_file=None, params="",
 
     return build_obj
     
-env.AddMethod(stata_run_params_factory, 'StataBuild')
 
 
-if GetOption("show_config"):	
-    SetOption("no_exec", True)
-    SetOption("silent", True)
-    SCons.Debug.Trace("Current config:\n", pre_sys_stdout)
-    if 'Programs' not in env['CONFIG'] or 'stata_exe' not in env['CONFIG']['Programs']:
-        SCons.Debug.Trace("Stata batch found automatically: "+env['STATABATCHCOM']+"\n", pre_sys_stdout)
-    for section in config.sections():
-        SCons.Debug.Trace("["+section+"]\n", pre_sys_stdout)
-        for key in config[section]:
-            SCons.Debug.Trace(key + ": " + config[section][key] + "\n", pre_sys_stdout)
-    SCons.Debug.Trace("\n", pre_sys_stdout)
+##############################################
+#                Info                        #
+##############################################
+AddOption(
+    '--show-config',
+    dest='show_config',
+    action='store_true',
+    default=False,
+    help='Show pystatacons configuration.',
+)
+
+
 
     
-
+##############################################
+#              Patching                      #
+##############################################
 # how to get a Stata-style datasignature
+int_env = None
 def get_datasign(fname):
-    m_str = env['CONFIG']['SCons']['use_custom_datasignature']
+    if 'STATABATCHEXE' not in int_env:
+        raise LookupError("Can't find Stata")
+    m_str = int_env['CONFIG']['SCons']['use_custom_datasignature']
     meta = m_str!="DataOnly" and m_str!="Datasignature"
     meta_arg_split = [] if meta else ["nometa"]
     vv_only = m_str=="VVLabelsOnly"
     vv_only_arg_split = ["vv_labels_only"] if vv_only else []
-    slow = ('Project' in env['CONFIG'] and 'cache_dir' in env['CONFIG']['Project']) or ('Project' in env['CONFIG'] and 'dta_sig_mode' in env['CONFIG']['Project'] and env['CONFIG']['Project']['dta_sig_mode']=='slow')
+    slow = ('Project' in int_env['CONFIG'] and 'cache_dir' in int_env['CONFIG']['Project']) or ('Project' in int_env['CONFIG'] and 'dta_sig_mode' in int_env['CONFIG']['Project'] and int_env['CONFIG']['Project']['dta_sig_mode']=='slow')
     fast_arg_split = [] if slow else ["fast"]
     fname_abs = os.path.abspath(fname)
     
     #Run in temp-dir as in parallel mode we don't want to processes to try writing to the same stata.log
     import tempfile
     with tempfile.TemporaryDirectory() as tmpdirname:
+        #tmpdirname = "."
         sig_fname="sig.txt"
-        args_split = [env['STATABATCHEXE'], env['STATABATCHFLAG'], 'complete_datasignature,', 'dta_file("' + fname_abs + '")', 'fname("' + sig_fname + '")'] + meta_arg_split + fast_arg_split + vv_only_arg_split
-        #SCons.Debug.Trace(str(args_split) + "\n", pre_sys_stdout) 
+        args_split = [int_env['STATABATCHEXE'], int_env['STATABATCHFLAG'], 'complete_datasignature,', 'dta_file("' + fname_abs + '")', 'fname("' + sig_fname + '")'] + meta_arg_split + fast_arg_split + vv_only_arg_split
         cproc = subprocess.run(args_split, cwd=tmpdirname)
         if cproc.returncode!=0: #In case the Stata executable has a real issue
             raise Exception("Couldn't get the file data-signature. Stata error")
@@ -372,6 +351,7 @@ def get_datasign(fname):
     return sig
 
 # replacement of hash_file_signature. Checks if .dta and then special call, otherwise hands off to reguarl hash_file_signature()
+special_sig_fns = {".dta": get_datasign}
 def hash_file_signature_new(fname, chunksize=65536, hash_format=None):
     """
     Generate the md5 signature of a file
@@ -387,13 +367,17 @@ def hash_file_signature_new(fname, chunksize=65536, hash_format=None):
     import SCons.Util
 
     #SCons.Debug.Trace('new hash: ' + fname + "\n", pre_sys_stdout) 
-    if os.path.splitext(fname)[1]==".dta":
+    fname_ext = os.path.splitext(fname)[1]
+    if fname_ext in special_sig_fns.keys():
         try:
-            sig = get_datasign(fname)
+            #SCons.Debug.Trace(str(fname_ext) + "\n", pre_sys_stdout)
+            sig = special_sig_fns[fname_ext](fname)
             if not GetOption('silent'):
-                SCons.Debug.Trace('Computed dta-signature: ' + fname + "\n", pre_sys_stdout)
+                SCons.Debug.Trace('Computed ' + fname_ext[1:] + '-signature: ' + fname + "\n", pre_sys_stdout)
             return sig
         except Exception as e:
+            print(e)
+            #SCons.Debug.Trace(str(fname_ext) + "\n", pre_sys_stdout)
             pass
 
     return SCons.Util.hash_file_signature(fname, chunksize, hash_format)
@@ -450,8 +434,13 @@ def get_csig_new(self) -> str:
 
 	return csig
 
-def monkey_patch_scons():
-    m_str = env['CONFIG']['SCons']['use_custom_datasignature']
+#Save the originals
+import SCons.Node.FS
+get_content_hash_orig = SCons.Node.FS.File.get_content_hash
+get_csig_orig = SCons.Node.FS.File.get_csig
+
+
+def monkey_patch_scons(m_str=""):
     if not GetOption('silent'):
         if m_str!="DataOnly" and m_str!="Datasignature" and m_str!="VVLabelsOnly":
             print("Using 'Strict' custom_datasignature.")
@@ -471,7 +460,6 @@ def monkey_patch_scons():
     ### Allow using Stata-style data-signatures instead of MD5 file hashes
     # Can't do this in SConstruct
     # Can't use Decider because I need my hash of the previous file and don't want to store it
-    import SCons.Node.FS
     # Can't easily replace SCons.Util.hash_file_signature because it's imported druing SCons/Script/__init__.py
     # (https://medium.com/@chipiga86/python-monkey-patching-like-a-boss-87d7ddb8098e)
     # So have to override the calling class method instead
@@ -479,7 +467,45 @@ def monkey_patch_scons():
     #Overwrite this one too that used a different hash method with small files
     SCons.Node.FS.File.get_csig = get_csig_new
 
-if env['CONFIG']['SCons']['use_custom_datasignature']!="False": monkey_patch_scons()
-elif not GetOption('silent'):
-    print("Using default timestamp-dependent checksums of dataset,")
-    print("Edit use_custom_datasignature in config_project.ini to change (Strict, DataOnly, VVLabelsOnly)")
+# Does: Package version check, initializes tool-less env (lots of stata users don't have any), reads config and adds to env, add Stata tool + builders to env, prints 'show-config', patches
+def init_env(env=None, patch_scons_sig_fns=True):
+    if packaging.version.parse(SCons.__version__) < packaging.version.parse("4.2.0"):
+        SCons.Debug.Trace("WARNING: You are running SCons version" + SCons.__version__ + " and statacons has only been tested on 4.2.0+. \n", pre_sys_stdout)
+
+    if env is None: env = DefaultEnvironment(ENV=os.environ, tools=[])
+    global int_env
+    int_env = env
+
+    if GetOption("config_user") is not None:
+        config = configuration(config_user_ini=GetOption("config_user"))
+    else:
+        config = configuration()
+
+    env['CONFIG'] = config
+
+    env.Tool(stata_tool)
+    
+    stata_do_bld = Builder(action=stata_run)
+    env.Append(BUILDERS={'StataDo': stata_do_bld})
+    env.AddMethod(stata_run_params_factory, 'StataBuild')
+    
+    if GetOption("show_config"):
+        SetOption("no_exec", True)
+        SetOption("silent", True)
+        print("Current config:")
+        if 'Programs' not in config or 'stata_exe' not in config['Programs']:
+            print("Stata batch found automatically: "+env['STATABATCHCOM'])
+        for section in config.sections():
+            print("["+section+"]")
+            for key in config[section]:
+                print(key + ": " + config[section][key])
+        print("\n")
+
+    if not GetOption("clean") and patch_scons_sig_fns:
+        if config['SCons']['use_custom_datasignature']!="False": 
+            monkey_patch_scons(config['SCons']['use_custom_datasignature'])
+        elif not GetOption('silent'):
+            print("Using default timestamp-dependent checksums of dataset,")
+            print("Edit use_custom_datasignature in config_project.ini to change (Strict, DataOnly, VVLabelsOnly)")
+
+    return(env)
